@@ -1,6 +1,6 @@
 /*
  * Remmina - The GTK+ Remote Desktop Client
- * Copyright (C) 2009 - Vic Lee 
+ * Copyright (C) 2009-2010 Vic Lee 
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -30,6 +30,8 @@
 #include "remminapref.h"
 #include "remminaprefdialog.h"
 #include "remminawidgetpool.h"
+#include "remminapluginmanager.h"
+#include "remminalog.h"
 #include "remminamain.h"
 
 G_DEFINE_TYPE (RemminaMain, remmina_main, GTK_TYPE_WINDOW)
@@ -38,10 +40,16 @@ struct _RemminaMainPriv
 {
     GtkWidget *file_list;
     GtkTreeModel *file_model;
+    GtkTreeModel *file_model_filter;
+    GtkTreeModel *file_model_sort;
     GtkUIManager *uimanager;
     GtkWidget *toolbar;
     GtkWidget *statusbar;
 
+    GtkToolItem *quick_search_separator;
+    GtkToolItem *quick_search_item;
+    GtkWidget *quick_search_entry;
+
     GtkTreeViewColumn *group_column;
 
     GtkActionGroup *main_group;
@@ -51,7 +59,6 @@ struct _RemminaMainPriv
 
     gchar *selected_filename;
     gchar *selected_name;
-    GtkTreeIter *selected_iter;
 };
 
 static void
@@ -69,6 +76,11 @@ enum
     N_COLUMNS
 };
 
+static GtkTargetEntry remmina_drop_types[] = 
+{
+    { "text/uri-list", 0, 1 }
+};
+
 static void
 remmina_main_save_size (RemminaMain *remminamain)
 {
@@ -170,11 +182,6 @@ remmina_main_load_file_list_callback (gpointer data, gpointer user_data)
         SERVER_COLUMN, remminafile->server,
         FILENAME_COLUMN, remminafile->filename,
         -1);
-
-    if (remminamain->priv->selected_filename && g_strcmp0 (remminafile->filename, remminamain->priv->selected_filename) == 0)
-    {
-        remminamain->priv->selected_iter = gtk_tree_iter_copy (&iter);
-    }
 }
 
 static void
@@ -239,11 +246,6 @@ remmina_main_load_file_tree_callback (gpointer data, gpointer user_data)
         SERVER_COLUMN, remminafile->server,
         FILENAME_COLUMN, remminafile->filename,
         -1);
-
-    if (remminamain->priv->selected_filename && g_strcmp0 (remminafile->filename, remminamain->priv->selected_filename) == 0)
-    {
-        remminamain->priv->selected_iter = gtk_tree_iter_copy (&child);
-    }
 }
 
 static void
@@ -258,13 +260,78 @@ remmina_main_file_model_on_sort (GtkTreeSortable *sortable, RemminaMain *remmina
     remmina_pref_save ();
 }
 
+static gboolean
+remmina_main_filter_visible_func (GtkTreeModel *model, GtkTreeIter *iter, RemminaMain *remminamain)
+{
+    gchar *text;
+    gchar *protocol, *name, *group, *server, *s;
+    gboolean result = TRUE;
+
+    if (!remmina_pref.show_quick_search) return TRUE;
+
+    text = g_ascii_strdown (gtk_entry_get_text (GTK_ENTRY (remminamain->priv->quick_search_entry)), -1);
+    if (text && text[0])
+    {
+        gtk_tree_model_get (model, iter,
+            PROTOCOL_COLUMN, &protocol,
+            NAME_COLUMN, &name,
+            GROUP_COLUMN, &group,
+            SERVER_COLUMN, &server,
+            -1);
+        if (g_strcmp0 (protocol, GTK_STOCK_DIRECTORY) != 0)
+        {
+            s = g_ascii_strdown (name, -1); g_free (name); name = s;
+            s = g_ascii_strdown (group, -1); g_free (group); group = s;
+            s = g_ascii_strdown (server, -1); g_free (server); server = s;
+            result = (strstr (name, text) || strstr (server, text) || strstr (group, text));
+        }
+        g_free (protocol);
+        g_free (name);
+        g_free (group);
+        g_free (server);
+    }
+    g_free (text);
+    return result;
+}
+
+static void
+remmina_main_select_file (RemminaMain *remminamain, const gchar *filename)
+{
+    GtkTreeIter iter;
+    GtkTreePath *path;
+    gchar *item_filename;
+    gboolean cmp;
+
+    if (!gtk_tree_model_get_iter_first (remminamain->priv->file_model_sort, &iter)) return;
+
+    while (1)
+    {
+        gtk_tree_model_get (remminamain->priv->file_model_sort, &iter, FILENAME_COLUMN, &item_filename, -1);
+        cmp = g_strcmp0 (item_filename, filename);
+        g_free (item_filename);
+        if (cmp == 0)
+        {
+            gtk_tree_selection_select_iter (
+                gtk_tree_view_get_selection (GTK_TREE_VIEW (remminamain->priv->file_list)),
+                &iter);
+            path = gtk_tree_model_get_path (remminamain->priv->file_model_sort, &iter);
+            gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (remminamain->priv->file_list),
+                path, NULL, TRUE, 0.5, 0.0);
+            gtk_tree_path_free (path);
+            return;
+        }
+        if (!gtk_tree_model_iter_next (remminamain->priv->file_model_sort, &iter)) return;
+    }
+}
+
 static void
 remmina_main_load_files (RemminaMain *remminamain)
 {
+    GtkTreeModel *filter;
+    GtkTreeModel *sort;
     gint n;
     gchar buf[200];
     guint context_id;
-    GtkTreePath *path;
 
     switch (remmina_pref.view_file_mode)
     {
@@ -286,24 +353,24 @@ remmina_main_load_files (RemminaMain *remminamain)
         break;
     }
 
-    gtk_tree_view_set_model (GTK_TREE_VIEW (remminamain->priv->file_list), remminamain->priv->file_model);
+    filter = gtk_tree_model_filter_new (remminamain->priv->file_model, NULL);
+    gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (filter),
+        (GtkTreeModelFilterVisibleFunc) remmina_main_filter_visible_func, remminamain, NULL);
+    remminamain->priv->file_model_filter = filter;
+
+    sort = gtk_tree_model_sort_new_with_model (filter);
+    remminamain->priv->file_model_sort = sort;
+
+    gtk_tree_view_set_model (GTK_TREE_VIEW (remminamain->priv->file_list), sort);
     gtk_tree_view_expand_all (GTK_TREE_VIEW (remminamain->priv->file_list));
-    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (remminamain->priv->file_model),
+    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (sort),
         remmina_pref.main_sort_column_id, remmina_pref.main_sort_order);
-    g_signal_connect (G_OBJECT (remminamain->priv->file_model), "sort-column-changed",
+    g_signal_connect (G_OBJECT (sort), "sort-column-changed",
         G_CALLBACK (remmina_main_file_model_on_sort), remminamain);
 
-    if (remminamain->priv->selected_iter)
+    if (remminamain->priv->selected_filename)
     {
-        gtk_tree_selection_select_iter (
-            gtk_tree_view_get_selection (GTK_TREE_VIEW (remminamain->priv->file_list)),
-            remminamain->priv->selected_iter);
-        path = gtk_tree_model_get_path (remminamain->priv->file_model, remminamain->priv->selected_iter);
-        gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (remminamain->priv->file_list),
-            path, NULL, TRUE, 0.5, 0.0);
-        gtk_tree_path_free (path);
-        gtk_tree_iter_free (remminamain->priv->selected_iter);
-        remminamain->priv->selected_iter = NULL;
+        remmina_main_select_file (remminamain, remminamain->priv->selected_filename);
     }
 
     g_snprintf (buf, 200, ngettext("Total %i item.", "Total %i items.", n), n);
@@ -313,7 +380,7 @@ remmina_main_load_files (RemminaMain *remminamain)
 }
 
 static void
-remmina_main_action_action_quick_connect_full (RemminaMain *remminamain, const gchar *protocol)
+remmina_main_action_action_quick_connect_full (const gchar *protocol)
 {
     GtkWidget *widget;
 
@@ -324,37 +391,13 @@ remmina_main_action_action_quick_connect_full (RemminaMain *remminamain, const g
 static void
 remmina_main_action_action_quick_connect (GObject *object, RemminaMain *remminamain)
 {
-    remmina_main_action_action_quick_connect_full (remminamain, NULL);
+    remmina_main_action_action_quick_connect_full (NULL);
 }
 
 static void
-remmina_main_action_action_quick_connect_rdp (GObject *object, RemminaMain *remminamain)
+remmina_main_action_action_quick_connect_proto (GtkWidget *widget, const gchar *protocol)
 {
-    remmina_main_action_action_quick_connect_full (remminamain, "RDP");
-}
-
-static void
-remmina_main_action_action_quick_connect_vnc (GObject *object, RemminaMain *remminamain)
-{
-    remmina_main_action_action_quick_connect_full (remminamain, "VNC");
-}
-
-static void
-remmina_main_action_action_quick_connect_vnci (GObject *object, RemminaMain *remminamain)
-{
-    remmina_main_action_action_quick_connect_full (remminamain, "VNCI");
-}
-
-static void
-remmina_main_action_action_quick_connect_xdmcp (GObject *object, RemminaMain *remminamain)
-{
-    remmina_main_action_action_quick_connect_full (remminamain, "XDMCP");
-}
-
-static void
-remmina_main_action_action_quick_connect_sftp (GObject *object, RemminaMain *remminamain)
-{
-    remmina_main_action_action_quick_connect_full (remminamain, "SFTP");
+    remmina_main_action_action_quick_connect_full (protocol);
 }
 
 static void
@@ -428,6 +471,7 @@ remmina_main_action_edit_delete (GtkAction *action, RemminaMain *remminamain)
         remmina_main_load_files (remminamain);
     }
     gtk_widget_destroy (dialog);
+    remmina_main_clear_selection_data (remminamain);
 }
 
 static void
@@ -448,9 +492,9 @@ remmina_main_action_action_quit (GtkAction *action, RemminaMain *remminamain)
 static void
 remmina_main_action_view_toolbar (GtkToggleAction *action, RemminaMain *remminamain)
 {
-	gboolean toggled;
+    gboolean toggled;
 
-	toggled = gtk_toggle_action_get_active (action);
+    toggled = gtk_toggle_action_get_active (action);
     if (toggled)
     {
         gtk_widget_show (remminamain->priv->toolbar);
@@ -466,12 +510,43 @@ remmina_main_action_view_toolbar (GtkToggleAction *action, RemminaMain *remminam
     }
 }
 
+static void
+remmina_main_action_view_quick_search (GtkToggleAction *action, RemminaMain *remminamain)
+{
+    gboolean toggled;
+
+    toggled = gtk_toggle_action_get_active (action);
+    if (toggled)
+    {
+        gtk_entry_set_text (GTK_ENTRY (remminamain->priv->quick_search_entry), "");
+        gtk_widget_show (GTK_WIDGET (remminamain->priv->quick_search_separator));
+        gtk_widget_show (GTK_WIDGET (remminamain->priv->quick_search_item));
+        gtk_widget_grab_focus (remminamain->priv->quick_search_entry);
+    }
+    else
+    {
+        gtk_widget_hide (GTK_WIDGET (remminamain->priv->quick_search_separator));
+        gtk_widget_hide (GTK_WIDGET (remminamain->priv->quick_search_item));
+    }
+    if (remminamain->priv->initialized)
+    {
+        remmina_pref.show_quick_search = toggled;
+        remmina_pref_save ();
+
+        if (!toggled)
+        {
+            gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (remminamain->priv->file_model_filter));
+            gtk_tree_view_expand_all (GTK_TREE_VIEW (remminamain->priv->file_list));
+        }
+    }
+}
+
 static void
 remmina_main_action_view_statusbar (GtkToggleAction *action, RemminaMain *remminamain)
 {
-	gboolean toggled;
+    gboolean toggled;
 
-	toggled = gtk_toggle_action_get_active (action);
+    toggled = gtk_toggle_action_get_active (action);
     if (toggled)
     {
         gtk_widget_show (remminamain->priv->statusbar);
@@ -490,9 +565,9 @@ remmina_main_action_view_statusbar (GtkToggleAction *action, RemminaMain *remmin
 static void
 remmina_main_action_view_small_toolbutton (GtkToggleAction *action, RemminaMain *remminamain)
 {
-	gboolean toggled;
+    gboolean toggled;
 
-	toggled = gtk_toggle_action_get_active (action);
+    toggled = gtk_toggle_action_get_active (action);
     if (toggled)
     {
         gtk_toolbar_set_icon_size (GTK_TOOLBAR (remminamain->priv->toolbar), GTK_ICON_SIZE_MENU);
@@ -516,6 +591,151 @@ remmina_main_action_view_file_mode (GtkRadioAction *action, GtkRadioAction *curr
     remmina_main_load_files (remminamain);
 }
 
+static void
+remmina_main_import_file_list (RemminaMain *remminamain, GSList *files)
+{
+    GtkWidget *dlg;
+    GSList *element;
+    gchar *path;
+    RemminaFilePlugin *plugin;
+    GString *err;
+    RemminaFile *remminafile = NULL;
+    gboolean imported;
+
+    err = g_string_new (NULL);
+    imported = FALSE;
+    for (element = files; element; element = element->next)
+    {
+        path = (gchar*) element->data;
+        plugin = remmina_plugin_manager_get_import_file_handler (path);
+        if (plugin && (remminafile = plugin->import_func (path)) != NULL &&
+            remminafile->name)
+        {
+            remminafile->filename = remmina_file_generate_filename ();
+            remmina_file_save_all (remminafile);
+            imported = TRUE;
+        }
+        else
+        {
+            g_string_append (err, path);
+            g_string_append_c (err, '\n');
+        }
+        if (remminafile)
+        {
+            remmina_file_free (remminafile);
+            remminafile = NULL;
+        }
+        g_free (path);
+    }
+    g_slist_free (files);
+    if (err->len > 0)
+    {
+        dlg = gtk_message_dialog_new (GTK_WINDOW (remminamain),
+            GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
+            _("Unable to import:\n%s"), err->str);
+        g_signal_connect (G_OBJECT (dlg), "response", G_CALLBACK (gtk_widget_destroy), NULL);
+        gtk_widget_show (dlg);
+    }
+    g_string_free (err, TRUE);
+    if (imported)
+    {
+        remmina_main_load_files (remminamain);
+    }
+}
+
+static void
+remmina_main_action_tools_import_on_response (GtkDialog *dialog, gint response_id, RemminaMain *remminamain)
+{
+    GSList *files;
+
+    if (response_id == GTK_RESPONSE_ACCEPT)
+    {
+        files = gtk_file_chooser_get_filenames (GTK_FILE_CHOOSER (dialog));
+        remmina_main_import_file_list (remminamain, files);
+    }
+    gtk_widget_destroy (GTK_WIDGET (dialog));
+}
+
+static void
+remmina_main_action_tools_import (GtkAction *action, RemminaMain *remminamain)
+{
+    GtkWidget *dialog;
+
+    dialog = gtk_file_chooser_dialog_new (_("Import"), GTK_WINDOW (remminamain),
+        GTK_FILE_CHOOSER_ACTION_OPEN,
+        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
+        NULL);
+    gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER (dialog), TRUE);
+    g_signal_connect (G_OBJECT (dialog), "response",
+        G_CALLBACK (remmina_main_action_tools_import_on_response), remminamain);
+    gtk_widget_show (dialog);
+}
+
+static void
+remmina_main_action_tools_export (GtkAction *action, RemminaMain *remminamain)
+{
+    RemminaFilePlugin *plugin;
+    RemminaFile *remminafile;
+    GtkWidget *dialog;
+
+    if (remminamain->priv->selected_filename == NULL) return;
+
+    remminafile = remmina_file_load (remminamain->priv->selected_filename);
+    if (remminafile == NULL) return;
+    plugin = remmina_plugin_manager_get_export_file_handler (remminafile);
+    if (plugin)
+    {
+        dialog = gtk_file_chooser_dialog_new (plugin->export_hints, GTK_WINDOW (remminamain),
+            GTK_FILE_CHOOSER_ACTION_SAVE,
+            GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
+            NULL);
+        if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
+        {
+            plugin->export_func (remminafile, gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog)));
+        }
+        gtk_widget_destroy (dialog);
+    }
+    else
+    {
+        dialog = gtk_message_dialog_new (GTK_WINDOW (remminamain),
+            GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
+            _("This protocol does not support exporting."));
+        g_signal_connect (G_OBJECT (dialog), "response", G_CALLBACK (gtk_widget_destroy), NULL);
+        gtk_widget_show (dialog);
+    }
+    remmina_file_free (remminafile);
+}
+
+static void
+remmina_main_action_tools_plugins (GtkAction *action, RemminaMain *remminamain)
+{
+    remmina_plugin_manager_show (GTK_WINDOW (remminamain));
+}
+
+static void
+remmina_main_action_tools_addition (GtkAction *action, RemminaMain *remminamain)
+{
+    RemminaToolPlugin *plugin;
+
+    plugin = (RemminaToolPlugin *) remmina_plugin_manager_get_plugin (REMMINA_PLUGIN_TYPE_TOOL, gtk_action_get_name (action));
+    if (plugin)
+    {
+        plugin->exec_func();
+    }
+}
+
+static void
+remmina_main_action_help_wiki (GtkAction *action, RemminaMain *remminamain)
+{
+    g_app_info_launch_default_for_uri ("http://sourceforge.net/apps/mediawiki/remmina/", NULL, NULL);
+}
+
+static void
+remmina_main_action_help_debug (GtkAction *action, RemminaMain *remminamain)
+{
+    remmina_log_start ();
+}
+
 static void
 remmina_main_action_help_about (GtkAction *action, RemminaMain *remminamain)
 {
@@ -527,17 +747,7 @@ static const gchar *remmina_main_ui_xml =
 "  <menubar name='MenuBar'>"
 "    <menu name='ActionMenu' action='Action'>"
 "      <menuitem name='ActionQuickConnectMenu' action='ActionQuickConnect'/>"
-"      <menu name='ActionQuickConnectProtoMenu' action='ActionQuickConnectProto'>"
-"        <menuitem name='ActionQuickConnectProtoRDPMenu' action='ActionQuickConnectProtoRDP'/>"
-#ifdef HAVE_LIBVNCCLIENT
-"        <menuitem name='ActionQuickConnectProtoVNCMenu' action='ActionQuickConnectProtoVNC'/>"
-"        <menuitem name='ActionQuickConnectProtoVNCIMenu' action='ActionQuickConnectProtoVNCI'/>"
-#endif
-"        <menuitem name='ActionQuickConnectProtoXDMCPMenu' action='ActionQuickConnectProtoXDMCP'/>"
-#ifdef HAVE_LIBSSH
-"        <menuitem name='ActionQuickConnectProtoSFTPMenu' action='ActionQuickConnectProtoSFTP'/>"
-#endif
-"      </menu>"
+"      <menuitem name='ActionQuickConnectProtoMenu' action='ActionQuickConnectProto'/>"
 "      <menuitem name='ActionConnectMenu' action='ActionConnect'/>"
 "      <separator/>"
 "      <menuitem name='ActionQuitMenu' action='ActionQuit'/>"
@@ -553,13 +763,24 @@ static const gchar *remmina_main_ui_xml =
 "    <menu name='ViewMenu' action='View'>"
 "      <menuitem name='ViewToolbarMenu' action='ViewToolbar'/>"
 "      <menuitem name='ViewStatusbarMenu' action='ViewStatusbar'/>"
+"      <menuitem name='ViewQuickSearchMenu' action='ViewQuickSearch'/>"
 "      <separator/>"
 "      <menuitem name='ViewSmallToolbuttonMenu' action='ViewSmallToolbutton'/>"
 "      <separator/>"
 "      <menuitem name='ViewFileListMenu' action='ViewFileList'/>"
 "      <menuitem name='ViewFileTreeMenu' action='ViewFileTree'/>"
 "    </menu>"
+"    <menu name='ToolsMenu' action='Tools'>"
+"      <menuitem name='ToolsImportMenu' action='ToolsImport'/>"
+"      <menuitem name='ToolsExportMenu' action='ToolsExport'/>"
+"      <placeholder name='ToolsAdditions'/>"
+"      <separator/>"
+"      <menuitem name='ToolsPluginsMenu' action='ToolsPlugins'/>"
+"    </menu>"
 "    <menu name='HelpMenu' action='Help'>"
+"      <menuitem name='HelpWikiMenu' action='HelpWiki'/>"
+"      <menuitem name='HelpDebugMenu' action='HelpDebug'/>"
+"      <separator/>"
 "      <menuitem name='HelpAboutMenu' action='HelpAbout'/>"
 "    </menu>"
 "  </menubar>"
@@ -573,17 +794,6 @@ static const gchar *remmina_main_ui_xml =
 "    <separator/>"
 "    <toolitem action='EditPreferences'/>"
 "  </toolbar>"
-"  <popup name='PopupQuickConnectProtoMenu' action='ActionQuickConnectProto'>"
-"    <menuitem name='PopupQuickConnectProtoRDPMenu' action='ActionQuickConnectProtoRDP'/>"
-#ifdef HAVE_LIBVNCCLIENT
-"    <menuitem name='PopupQuickConnectProtoVNCMenu' action='ActionQuickConnectProtoVNC'/>"
-"    <menuitem name='PopupQuickConnectProtoVNCIMenu' action='ActionQuickConnectProtoVNCI'/>"
-#endif
-"    <menuitem name='PopupQuickConnectProtoXDMCPMenu' action='ActionQuickConnectProtoXDMCP'/>"
-#ifdef HAVE_LIBSSH
-"    <menuitem name='PopupQuickConnectProtoSFTPMenu' action='ActionQuickConnectProtoSFTP'/>"
-#endif
-"  </popup>"
 "  <popup name='PopupMenu'>"
 "    <menuitem action='ActionConnect'/>"
 "    <separator/>"
@@ -598,6 +808,7 @@ static const GtkActionEntry remmina_main_ui_menu_entries[] =
     { "Action", NULL, N_("_Action") },
     { "Edit", NULL, N_("_Edit") },
     { "View", NULL, N_("_View") },
+    { "Tools", NULL, N_("_Tools") },
     { "Help", NULL, N_("_Help") },
 
     { "ActionQuickConnect", GTK_STOCK_JUMP_TO, N_("Quick Connect"), "<control>U",
@@ -606,17 +817,6 @@ static const GtkActionEntry remmina_main_ui_menu_entries[] =
 
     { "ActionQuickConnectProto", NULL, N_("Quick _Connect To") },
 
-    { "ActionQuickConnectProtoRDP", "remmina-rdp", N_("RDP - Windows Terminal Service"), NULL, NULL,
-        G_CALLBACK (remmina_main_action_action_quick_connect_rdp) },
-    { "ActionQuickConnectProtoVNC", "remmina-vnc", N_("VNC - Virtual Network Computing"), NULL, NULL,
-        G_CALLBACK (remmina_main_action_action_quick_connect_vnc) },
-    { "ActionQuickConnectProtoVNCI", "remmina-vnc", N_("VNC - Incoming Connection"), NULL, NULL,
-        G_CALLBACK (remmina_main_action_action_quick_connect_vnci) },
-    { "ActionQuickConnectProtoXDMCP", "remmina-xdmcp", N_("XDMCP - X Remote Session"), NULL, NULL,
-        G_CALLBACK (remmina_main_action_action_quick_connect_xdmcp) },
-    { "ActionQuickConnectProtoSFTP", "remmina-sftp", N_("SFTP - Secure File Transfer"), NULL, NULL,
-        G_CALLBACK (remmina_main_action_action_quick_connect_sftp) },
-
     { "EditNew", GTK_STOCK_NEW, NULL, "<control>N",
         N_("Create a new remote desktop file"),
         G_CALLBACK (remmina_main_action_edit_new) },
@@ -629,6 +829,22 @@ static const GtkActionEntry remmina_main_ui_menu_entries[] =
         N_("Quit Remmina"),
         G_CALLBACK (remmina_main_action_action_quit) },
 
+    { "ToolsImport", NULL, N_("Import"), NULL,
+        NULL,
+        G_CALLBACK (remmina_main_action_tools_import) },
+
+    { "ToolsPlugins", NULL, N_("Plugins"), NULL,
+        NULL,
+        G_CALLBACK (remmina_main_action_tools_plugins) },
+
+    { "HelpWiki", NULL, N_("Online Wiki"), NULL,
+        NULL,
+        G_CALLBACK (remmina_main_action_help_wiki) },
+
+    { "HelpDebug", NULL, N_("Debug Window"), NULL,
+        NULL,
+        G_CALLBACK (remmina_main_action_help_debug) },
+
     { "HelpAbout", GTK_STOCK_ABOUT, NULL, NULL,
         NULL,
         G_CALLBACK (remmina_main_action_help_about) }
@@ -650,7 +866,11 @@ static const GtkActionEntry remmina_main_ui_file_sensitive_menu_entries[] =
 
     { "EditDelete", GTK_STOCK_DELETE, NULL, "<control>D",
         N_("Delete the selected remote desktop file"),
-        G_CALLBACK (remmina_main_action_edit_delete) }
+        G_CALLBACK (remmina_main_action_edit_delete) },
+
+    { "ToolsExport", NULL, N_("Export"), NULL,
+        NULL,
+        G_CALLBACK (remmina_main_action_tools_export) }
 };
 
 static const GtkToggleActionEntry remmina_main_ui_toggle_menu_entries[] =
@@ -661,6 +881,9 @@ static const GtkToggleActionEntry remmina_main_ui_toggle_menu_entries[] =
     { "ViewStatusbar", NULL, N_("Statusbar"), NULL, NULL,
         G_CALLBACK (remmina_main_action_view_statusbar), TRUE },
 
+    { "ViewQuickSearch", NULL, N_("Quick Search"), NULL, NULL,
+        G_CALLBACK (remmina_main_action_view_quick_search), FALSE },
+
     { "ViewSmallToolbutton", NULL, N_("Small Toolbar Buttons"), NULL, NULL,
         G_CALLBACK (remmina_main_action_view_small_toolbutton), FALSE }
 };
@@ -700,12 +923,122 @@ remmina_main_file_list_on_button_press (GtkWidget *widget, GdkEventButton *event
     return FALSE;
 }
 
+static gboolean
+remmina_main_iterate_protocol_menu (gchar *protocol, RemminaPlugin *plugin, gpointer data)
+{
+    GtkWidget *item;
+    GtkWidget *image;
+
+    item = gtk_image_menu_item_new_with_label (plugin->description);
+    gtk_widget_show (item);
+
+    image = gtk_image_new_from_icon_name (((RemminaProtocolPlugin *)plugin)->icon_name, GTK_ICON_SIZE_MENU);
+    gtk_widget_show (image);
+    gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
+    gtk_image_menu_item_set_always_show_image (GTK_IMAGE_MENU_ITEM (item), TRUE);
+
+    gtk_menu_shell_append (GTK_MENU_SHELL (data), item);
+
+    g_signal_connect (G_OBJECT (item), "activate",
+        G_CALLBACK (remmina_main_action_action_quick_connect_proto), (gpointer) plugin->name);
+
+    return FALSE;
+}
+
+static void
+remmina_main_quick_search_on_icon_press (GtkEntry *entry,
+    GtkEntryIconPosition icon_pos,
+    GdkEvent *event,
+    RemminaMain *remminamain)
+{
+    if (icon_pos == GTK_ENTRY_ICON_SECONDARY)
+    {
+        gtk_entry_set_text (entry, "");
+    }
+}
+
+static void
+remmina_main_quick_search_on_changed (GtkEditable *editable, RemminaMain *remminamain)
+{
+    gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (remminamain->priv->file_model_filter));
+    gtk_tree_view_expand_all (GTK_TREE_VIEW (remminamain->priv->file_list));
+}
+
+static void
+remmina_main_create_quick_search (RemminaMain *remminamain)
+{
+    GtkWidget *widget;
+    GValue val = { 0 };
+
+    remminamain->priv->quick_search_separator = gtk_separator_tool_item_new ();
+    gtk_toolbar_insert (GTK_TOOLBAR (remminamain->priv->toolbar), remminamain->priv->quick_search_separator, -1);
+
+    remminamain->priv->quick_search_item = gtk_tool_item_new ();
+    gtk_toolbar_insert (GTK_TOOLBAR (remminamain->priv->toolbar), remminamain->priv->quick_search_item, -1);
+
+    widget = gtk_entry_new ();
+    gtk_widget_show (widget);
+    gtk_entry_set_icon_from_stock (GTK_ENTRY (widget), GTK_ENTRY_ICON_PRIMARY, GTK_STOCK_FIND);
+    gtk_entry_set_icon_from_stock (GTK_ENTRY (widget), GTK_ENTRY_ICON_SECONDARY, GTK_STOCK_CLEAR);
+    gtk_entry_set_width_chars (GTK_ENTRY (widget), 25);
+    gtk_container_add (GTK_CONTAINER (remminamain->priv->quick_search_item), widget);
+
+    g_value_init (&val, G_TYPE_BOOLEAN);
+    g_value_set_boolean (&val, FALSE);
+    g_object_set_property (G_OBJECT (widget), "primary-icon-activatable", &val);
+    g_value_unset (&val);
+
+    g_signal_connect (G_OBJECT (widget), "icon-press", G_CALLBACK (remmina_main_quick_search_on_icon_press), remminamain);
+    g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (remmina_main_quick_search_on_changed), remminamain);
+
+    remminamain->priv->quick_search_entry = widget;
+}
+
+static void
+remmina_main_on_drag_data_received (RemminaMain *remminamain, GdkDragContext *drag_context, gint x, gint y,
+    GtkSelectionData *data, guint info, guint time, gpointer user_data)
+{
+    gchar **uris;
+    GSList *files = NULL;
+    gint i;
+
+    uris = g_uri_list_extract_uris ((gchar *) data->data);
+    for (i = 0; uris[i]; i++)
+    {
+        if (strncmp (uris[i], "file://", 7) != 0) continue;
+        files = g_slist_append (files, g_strdup (uris[i] + 7));
+    }
+    g_strfreev (uris);
+    remmina_main_import_file_list (remminamain, files);
+}
+
+static gboolean
+remmina_main_add_tool_plugin (gchar *name, RemminaPlugin *plugin, gpointer data)
+{
+    RemminaMain *remminamain = REMMINA_MAIN (data);
+    guint merge_id;
+    GtkAction *action;
+
+    merge_id = gtk_ui_manager_new_merge_id (remminamain->priv->uimanager);
+    action = gtk_action_new (name, plugin->description, NULL, NULL);
+    gtk_action_group_add_action (remminamain->priv->main_group, action);
+    g_signal_connect (G_OBJECT (action), "activate", G_CALLBACK (remmina_main_action_tools_addition), remminamain);
+    g_object_unref (action);
+
+    gtk_ui_manager_add_ui (remminamain->priv->uimanager, merge_id, "/MenuBar/ToolsMenu/ToolsAdditions",
+        name, name, GTK_UI_MANAGER_MENUITEM, FALSE);
+
+    return FALSE;
+}
+
 static void
 remmina_main_init (RemminaMain *remminamain)
 {
     RemminaMainPriv *priv;
     GtkWidget *vbox;
     GtkWidget *menubar;
+    GtkWidget *menu;
+    GtkWidget *menuitem;
     GtkUIManager *uimanager;
     GtkActionGroup *action_group;
     GtkWidget *scrolledwindow;
@@ -721,7 +1054,6 @@ remmina_main_init (RemminaMain *remminamain)
     priv->initialized = FALSE;
     priv->selected_filename = NULL;
     priv->selected_name = NULL;
-    priv->selected_iter = NULL;
 
     /* Create main window */
     g_signal_connect (G_OBJECT (remminamain), "delete-event", G_CALLBACK (remmina_main_on_delete_event), NULL);
@@ -775,12 +1107,16 @@ remmina_main_init (RemminaMain *remminamain)
         g_error_free (error);
     }
 
+    remmina_plugin_manager_for_each_plugin (REMMINA_PLUGIN_TYPE_TOOL, remmina_main_add_tool_plugin, remminamain);
+
     menubar = gtk_ui_manager_get_widget (uimanager, "/MenuBar");
     gtk_box_pack_start (GTK_BOX (vbox), menubar, FALSE, FALSE, 0);
 
     priv->toolbar = gtk_ui_manager_get_widget (uimanager, "/ToolBar");
     gtk_box_pack_start (GTK_BOX (vbox), priv->toolbar, FALSE, FALSE, 0);
 
+    remmina_main_create_quick_search (remminamain);
+
     /* Customized menu items and toolbar items which can't be built through standard ui manager */
     toolitem = gtk_menu_tool_button_new_from_stock (GTK_STOCK_JUMP_TO);
     gtk_widget_show (GTK_WIDGET (toolitem));
@@ -789,8 +1125,14 @@ remmina_main_init (RemminaMain *remminamain)
     g_signal_connect (G_OBJECT (toolitem), "clicked",
         G_CALLBACK (remmina_main_action_action_quick_connect), remminamain);
 
-    gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON (toolitem),
-        gtk_ui_manager_get_widget (uimanager, "/PopupQuickConnectProtoMenu"));
+    menu = gtk_menu_new ();
+    remmina_plugin_manager_for_each_plugin (REMMINA_PLUGIN_TYPE_PROTOCOL, remmina_main_iterate_protocol_menu, menu);
+    gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON (toolitem), menu);
+
+    menuitem = gtk_ui_manager_get_widget (uimanager, "/MenuBar/ActionMenu/ActionQuickConnectProtoMenu");
+    menu = gtk_menu_new ();
+    remmina_plugin_manager_for_each_plugin (REMMINA_PLUGIN_TYPE_PROTOCOL, remmina_main_iterate_protocol_menu, menu);
+    gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);
 
     gtk_window_add_accel_group (GTK_WINDOW (remminamain), gtk_ui_manager_get_accel_group (uimanager));
 
@@ -864,12 +1206,22 @@ remmina_main_init (RemminaMain *remminamain)
         gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (
             gtk_action_group_get_action (priv->main_group, "ViewStatusbar")), FALSE);
     }
+    if (remmina_pref.show_quick_search)
+    {
+        gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (
+            gtk_action_group_get_action (priv->main_group, "ViewQuickSearch")), TRUE);
+    }
     if (remmina_pref.small_toolbutton)
     {
         gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (
             gtk_action_group_get_action (priv->main_group, "ViewSmallToolbutton")), TRUE);
     }
 
+    /* Drag-n-drop support */
+    gtk_drag_dest_set (GTK_WIDGET (remminamain), GTK_DEST_DEFAULT_ALL, remmina_drop_types, 1, GDK_ACTION_COPY);
+    g_signal_connect (G_OBJECT (remminamain), "drag-data-received",
+        G_CALLBACK (remmina_main_on_drag_data_received), NULL);
+
     priv->initialized = TRUE;
 
     remmina_widget_pool_register (GTK_WIDGET (remminamain));
