@@ -1,6 +1,6 @@
 /*
  * Remmina - The GTK+ Remote Desktop Client
- * Copyright (C) 2009 - Vic Lee 
+ * Copyright (C) 2009-2010 Vic Lee 
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,6 +24,7 @@
 #include "remminapublic.h"
 #include "remminastringlist.h"
 #include "remminawidgetpool.h"
+#include "remminakeychooser.h"
 #include "remminapref.h"
 #include "remminaprefdialog.h"
 
@@ -67,7 +68,10 @@ struct _RemminaPrefDialogPriv
 {
     GtkWidget *notebook;
     GtkWidget *save_view_mode_check;
+    GtkWidget *save_when_connect_check;
     GtkWidget *invisible_toolbar_check;
+    GtkWidget *always_show_tab_check;
+    GtkWidget *hide_connection_toolbar_check;
     GtkWidget *default_action_combo;
     GtkWidget *default_mode_combo;
     GtkWidget *tab_mode_combo;
@@ -77,6 +81,16 @@ struct _RemminaPrefDialogPriv
     GtkWidget *resolutions_list;
     GtkWidget *applet_quick_ontop_check;
     GtkWidget *applet_hide_count_check;
+    GtkWidget *hostkey_chooser;
+    GtkWidget *shortcutkey_fullscreen_chooser;
+    GtkWidget *shortcutkey_autofit_chooser;
+    GtkWidget *shortcutkey_nexttab_chooser;
+    GtkWidget *shortcutkey_prevtab_chooser;
+    GtkWidget *shortcutkey_scale_chooser;
+    GtkWidget *shortcutkey_grab_chooser;
+    GtkWidget *shortcutkey_minimize_chooser;
+    GtkWidget *shortcutkey_disconnect_chooser;
+    GtkWidget *shortcutkey_toolbar_chooser;
 };
 
 static void
@@ -156,7 +170,10 @@ remmina_pref_dialog_destroy (GtkWidget *widget, gpointer data)
     RemminaPrefDialogPriv *priv = REMMINA_PREF_DIALOG (widget)->priv;
 
     remmina_pref.save_view_mode = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->save_view_mode_check));
+    remmina_pref.save_when_connect = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->save_when_connect_check));
     remmina_pref.invisible_toolbar = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->invisible_toolbar_check));
+    remmina_pref.always_show_tab = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->always_show_tab_check));
+    remmina_pref.hide_connection_toolbar = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->hide_connection_toolbar_check));
 
     s = gtk_combo_box_get_active_text (GTK_COMBO_BOX (priv->default_action_combo));
     remmina_pref.default_action = atoi (s);
@@ -185,6 +202,17 @@ remmina_pref_dialog_destroy (GtkWidget *widget, gpointer data)
     remmina_pref.applet_quick_ontop = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->applet_quick_ontop_check));
     remmina_pref.applet_hide_count = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->applet_hide_count_check));
 
+    remmina_pref.hostkey = REMMINA_KEY_CHOOSER (priv->hostkey_chooser)->keyval;
+    remmina_pref.shortcutkey_fullscreen = REMMINA_KEY_CHOOSER (priv->shortcutkey_fullscreen_chooser)->keyval;
+    remmina_pref.shortcutkey_autofit = REMMINA_KEY_CHOOSER (priv->shortcutkey_autofit_chooser)->keyval;
+    remmina_pref.shortcutkey_nexttab = REMMINA_KEY_CHOOSER (priv->shortcutkey_nexttab_chooser)->keyval;
+    remmina_pref.shortcutkey_prevtab = REMMINA_KEY_CHOOSER (priv->shortcutkey_prevtab_chooser)->keyval;
+    remmina_pref.shortcutkey_scale = REMMINA_KEY_CHOOSER (priv->shortcutkey_scale_chooser)->keyval;
+    remmina_pref.shortcutkey_grab = REMMINA_KEY_CHOOSER (priv->shortcutkey_grab_chooser)->keyval;
+    remmina_pref.shortcutkey_minimize = REMMINA_KEY_CHOOSER (priv->shortcutkey_minimize_chooser)->keyval;
+    remmina_pref.shortcutkey_disconnect = REMMINA_KEY_CHOOSER (priv->shortcutkey_disconnect_chooser)->keyval;
+    remmina_pref.shortcutkey_toolbar = REMMINA_KEY_CHOOSER (priv->shortcutkey_toolbar_chooser)->keyval;
+
     remmina_pref_save ();
     g_free (priv);
 }
@@ -230,7 +258,7 @@ remmina_pref_dialog_init (RemminaPrefDialog *dialog)
     gtk_widget_show (vbox);
     gtk_notebook_append_page (GTK_NOTEBOOK (notebook), vbox, tablabel);
 
-    table = gtk_table_new (8, 2, FALSE);
+    table = gtk_table_new (11, 2, FALSE);
     gtk_widget_show (table);
     gtk_table_set_row_spacings (GTK_TABLE (table), 4);
     gtk_table_set_col_spacings (GTK_TABLE (table), 4);
@@ -243,9 +271,15 @@ remmina_pref_dialog_init (RemminaPrefDialog *dialog)
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), remmina_pref.save_view_mode);
     priv->save_view_mode_check = widget;
 
-    widget = gtk_check_button_new_with_label (_("Invisible Floating Toolbar"));
+    widget = gtk_check_button_new_with_label (_("Save Settings When Starting the Connection"));
     gtk_widget_show (widget);
     gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 2, 1, 2);
+    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), remmina_pref.save_when_connect);
+    priv->save_when_connect_check = widget;
+
+    widget = gtk_check_button_new_with_label (_("Invisible Floating Toolbar"));
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 2, 2, 3);
     if (gtk_widget_is_composited (GTK_WIDGET (dialog)))
     {
         gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), remmina_pref.invisible_toolbar);
@@ -256,54 +290,66 @@ remmina_pref_dialog_init (RemminaPrefDialog *dialog)
     }
     priv->invisible_toolbar_check = widget;
 
+    widget = gtk_check_button_new_with_label (_("Always Show Tabs"));
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 2, 3, 4);
+    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), remmina_pref.always_show_tab);
+    priv->always_show_tab_check = widget;
+
+    widget = gtk_check_button_new_with_label (_("Hide Connection Toolbar"));
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 2, 4, 5);
+    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), remmina_pref.hide_connection_toolbar);
+    priv->hide_connection_toolbar_check = widget;
+
     widget = gtk_label_new (_("Double-click Action"));
     gtk_widget_show (widget);
     gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 2, 3);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 5, 6);
 
     widget = remmina_public_create_combo_mapint (default_action_list, remmina_pref.default_action, FALSE);
     gtk_widget_show (widget);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 2, 3);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 5, 6);
     priv->default_action_combo = widget;
 
     widget = gtk_label_new (_("Default View Mode"));
     gtk_widget_show (widget);
     gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 3, 4);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 6, 7);
 
     widget = remmina_public_create_combo_mapint (default_mode_list, remmina_pref.default_mode, FALSE);
     gtk_widget_show (widget);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 3, 4);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 6, 7);
     priv->default_mode_combo = widget;
 
     widget = gtk_label_new (_("Tab Interface"));
     gtk_widget_show (widget);
     gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 4, 5);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 7, 8);
 
     widget = remmina_public_create_combo_mapint (tab_mode_list, remmina_pref.tab_mode, FALSE);
     gtk_widget_show (widget);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 4, 5);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 7, 8);
     priv->tab_mode_combo = widget;
 
     widget = gtk_label_new (_("Scale Quality"));
     gtk_widget_show (widget);
     gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 5, 6);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 8, 9);
 
     widget = remmina_public_create_combo_mapint (scale_quality_list, remmina_pref.scale_quality, FALSE);
     gtk_widget_show (widget);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 5, 6);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 8, 9);
     priv->scale_quality_combo = widget;
 
     widget = gtk_label_new (_("SSH Tunnel Local Port"));
     gtk_widget_show (widget);
     gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 6, 7);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 9, 10);
 
     widget = gtk_entry_new_with_max_length (5);
     gtk_widget_show (widget);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 6, 7);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 9, 10);
     g_snprintf (buf, sizeof (buf), "%i", remmina_pref.sshtunnel_port);
     gtk_entry_set_text (GTK_ENTRY (widget), buf);
     priv->sshtunnel_port_entry = widget;
@@ -311,11 +357,11 @@ remmina_pref_dialog_init (RemminaPrefDialog *dialog)
     widget = gtk_label_new (_("Maximum Recent Items"));
     gtk_widget_show (widget);
     gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
-    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 7, 8);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 0, 1, 10, 11);
 
     hbox = gtk_hbox_new (FALSE, 2);
     gtk_widget_show (hbox);
-    gtk_table_attach_defaults (GTK_TABLE (table), hbox, 1, 2, 7, 8);
+    gtk_table_attach_defaults (GTK_TABLE (table), hbox, 1, 2, 10, 11);
 
     widget = gtk_entry_new_with_max_length (2);
     gtk_widget_show (widget);
@@ -377,6 +423,121 @@ remmina_pref_dialog_init (RemminaPrefDialog *dialog)
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), remmina_pref.applet_hide_count);
     priv->applet_hide_count_check = widget;
 
+    /* Keyboard tab */
+    tablabel = gtk_label_new (_("Keyboard"));
+    gtk_widget_show (tablabel);
+
+    /* Keyboard body */
+    vbox = gtk_vbox_new (FALSE, 0);
+    gtk_widget_show (vbox);
+    gtk_notebook_append_page (GTK_NOTEBOOK (notebook), vbox, tablabel);
+
+    table = gtk_table_new (9, 2, FALSE);
+    gtk_widget_show (table);
+    gtk_table_set_row_spacings (GTK_TABLE (table), 4);
+    gtk_table_set_col_spacings (GTK_TABLE (table), 4);
+    gtk_container_set_border_width (GTK_CONTAINER (table), 8);
+    gtk_box_pack_start (GTK_BOX (vbox), table, FALSE, FALSE, 0);
+
+    widget = gtk_label_new (_("Host Key"));
+    gtk_widget_show (widget);
+    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+    gtk_table_attach (GTK_TABLE (table), widget, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
+
+    widget = remmina_key_chooser_new (remmina_pref.hostkey);
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 0, 1);
+    priv->hostkey_chooser = widget;
+
+    widget = gtk_label_new (_("Toggle Fullscreen Mode"));
+    gtk_widget_show (widget);
+    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+    gtk_table_attach (GTK_TABLE (table), widget, 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
+
+    widget = remmina_key_chooser_new (remmina_pref.shortcutkey_fullscreen);
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 1, 2);
+    priv->shortcutkey_fullscreen_chooser = widget;
+
+    widget = gtk_label_new (_("Auto-Fit Window"));
+    gtk_widget_show (widget);
+    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+    gtk_table_attach (GTK_TABLE (table), widget, 0, 1, 2, 3, GTK_FILL, 0, 0, 0);
+
+    widget = remmina_key_chooser_new (remmina_pref.shortcutkey_autofit);
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 2, 3);
+    priv->shortcutkey_autofit_chooser = widget;
+
+    widget = gtk_label_new (_("Switch Tab Pages"));
+    gtk_widget_show (widget);
+    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+    gtk_table_attach (GTK_TABLE (table), widget, 0, 1, 3, 4, GTK_FILL, 0, 0, 0);
+
+    hbox = gtk_hbox_new (TRUE, 2);
+    gtk_widget_show (hbox);
+    gtk_table_attach_defaults (GTK_TABLE (table), hbox, 1, 2, 3, 4);
+
+    widget = remmina_key_chooser_new (remmina_pref.shortcutkey_prevtab);
+    gtk_widget_show (widget);
+    gtk_box_pack_start (GTK_BOX (hbox), widget, TRUE, TRUE, 0);
+    priv->shortcutkey_prevtab_chooser = widget;
+
+    widget = remmina_key_chooser_new (remmina_pref.shortcutkey_nexttab);
+    gtk_widget_show (widget);
+    gtk_box_pack_start (GTK_BOX (hbox), widget, TRUE, TRUE, 0);
+    priv->shortcutkey_nexttab_chooser = widget;
+
+    widget = gtk_label_new (_("Toggle Scaled Mode"));
+    gtk_widget_show (widget);
+    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+    gtk_table_attach (GTK_TABLE (table), widget, 0, 1, 4, 5, GTK_FILL, 0, 0, 0);
+
+    widget = remmina_key_chooser_new (remmina_pref.shortcutkey_scale);
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 4, 5);
+    priv->shortcutkey_scale_chooser = widget;
+
+    widget = gtk_label_new (_("Grab Keyboard"));
+    gtk_widget_show (widget);
+    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+    gtk_table_attach (GTK_TABLE (table), widget, 0, 1, 5, 6, GTK_FILL, 0, 0, 0);
+
+    widget = remmina_key_chooser_new (remmina_pref.shortcutkey_grab);
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 5, 6);
+    priv->shortcutkey_grab_chooser = widget;
+
+    widget = gtk_label_new (_("Minimize Window"));
+    gtk_widget_show (widget);
+    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+    gtk_table_attach (GTK_TABLE (table), widget, 0, 1, 6, 7, GTK_FILL, 0, 0, 0);
+
+    widget = remmina_key_chooser_new (remmina_pref.shortcutkey_minimize);
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 6, 7);
+    priv->shortcutkey_minimize_chooser = widget;
+
+    widget = gtk_label_new (_("Disconnect"));
+    gtk_widget_show (widget);
+    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+    gtk_table_attach (GTK_TABLE (table), widget, 0, 1, 7, 8, GTK_FILL, 0, 0, 0);
+
+    widget = remmina_key_chooser_new (remmina_pref.shortcutkey_disconnect);
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 7, 8);
+    priv->shortcutkey_disconnect_chooser = widget;
+
+    widget = gtk_label_new (_("Show / Hide Toolbar"));
+    gtk_widget_show (widget);
+    gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+    gtk_table_attach (GTK_TABLE (table), widget, 0, 1, 8, 9, GTK_FILL, 0, 0, 0);
+
+    widget = remmina_key_chooser_new (remmina_pref.shortcutkey_toolbar);
+    gtk_widget_show (widget);
+    gtk_table_attach_defaults (GTK_TABLE (table), widget, 1, 2, 8, 9);
+    priv->shortcutkey_toolbar_chooser = widget;
+
     remmina_widget_pool_register (GTK_WIDGET (dialog));
 }
 
